/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 *
 *  conn.c
 *  
 *
 */

#include "jserver.h"

/*
 * Structure: namespaces
 * This is a list of XML namespaces and the the connection handler
 * that should be used to handle incoming connections using that namespace
 */
struct namespaces
{
    char *ns;
    xmlstream_onNode handler;
    struct namespaces *next;
};


/*
 *  js_conn_namespace -- find the connection handler for a namespace
 *  Allows connection handlers to be associated with namespaces, and 
 *  then retreived when a connection on that namespace is recieved
 *
 *  parameters
 *      cmd -- C_SET to register a handler for a namespace
 *             C_CHECK to retreive a previously set handler
 *
 *      ns -- a string containing an XML namespace
 *
 *      handler -- a function pointer to the handler 
 *
 *  returns
 *      a function pointer to the handler if the command was sucessful 
 *      NULL if it was unsuccessful
 */
xmlstream_onNode js_conn_namespace(command cmd, char *ns, xmlstream_onNode handler)
{

    static struct namespaces *n = NULL;     /* list of handlers */
    struct namespaces *next;                /* used traverse the list */

    /* return NULL if no namespace is specified */
    if(ns == NULL)
        return NULL;


    /* handle the various commands */
    switch(cmd)
    {

        /* saves this namespace/handler pair */
    case C_SET:

        /* make sure there isn't already a handler registered */
        if(js_conn_namespace(C_CHECK,ns,NULL) != NULL)
            return NULL;

        /* create the list node */
        next = malloc(sizeof(struct namespaces));
        next->ns = j_strdup(ns);
        next->handler = handler;

        /* push it on to the list head */
        next->next = n;
        n = next;

        /* report success */
        return handler;
        break;

        /* returns the handler for the given namespace or returns NULL */
    case C_CHECK:

        /* start at the top of the list */
        next = n;
        while(next != NULL)
        {

            /* return the handler if the namespaces match */
            if(j_strcmp(ns,next->ns) == 0)
                return next->handler;

            /* no match - go on to the next namespace */
            next = next->next;

        }
        break;
    default:
    }

    /* if we got here there was no match */
    return NULL;

}


/*
 *  js_conn_connect -- handle new connection
 *  This is the callback on the listener thread whenever there is a new connect
 *  All it does is create an xmlstream to read data from the socket
 *
 *  parameters
 *      t -- not used, but required for all tlisten callbacks
 *      sock -- the socket with the incoming connection
 *      sa -- the sock_addr structure for the incoming connection
 *
 */
void js_conn_connect(thread t, int sock, struct sockaddr_in sa)
{
    thread new;
    /*    log_debug(ZONE,"jserver connection established (fd: %d, ip: %s, port: %d)\n",
                    sock, inet_ntoa(sa.sin_addr), ntohs(sa.sin_port));*/

    new = xmlstream_new(sock, inet_ntoa(sa.sin_addr), (xmlstream_onNode)js_conn_negotiate);

}


/*
 *  js_conn_negotiate -- negotiate an XML stream open tag
 *  This function is called whenever there is a new connect and root node is received. 
 *  It handles errors in the stream setup, then hands off to the correct handler for
 *  the stream's XML namespace.
 *  
 *  parameters
 *      t -- the thread which is handling the stream
 *      in -- the root node of the XML stream
 *      type -- the the node type of in (#defined in etherx.h)
 *
 */
void js_conn_negotiate(thread t, xmlnode in, int type)
{

    char *ns, *to;                      /* attributes of the root tag */
    tstream ts = (tstream)(t->data);    /* pointer to the thread's stream */
    xmlstream_onNode handler;           /* pointer to a node handler for the stream */

    /* ignore non-root nodes - this shouldn't get called for other node types */
    if(type != XMLSTREAM_ROOT)
        return;

    log_debug(ZONE,"New XML Stream started, locating internal namespace handler");

    /* get the namespace and hostname from the root stream node */
    ns = xmlnode_get_attrib(in, "xmlns");
    to = xmlnode_get_attrib(in, "to");

    /* find the handler for this namespace */
    handler = js_conn_namespace(C_CHECK,ns,NULL);

    /* is there an error in the stream tag? */
    if(type == XMLSTREAM_ERROR || ns == NULL || to == NULL || handler == NULL || !js_config_name(C_CHECK,to))
    {
        /* error! return an error message to the client */
        xmlstream_write_header(t, xmlstream_header(NULL, NULL, NULL));
        if(type == XMLSTREAM_ERROR)
            xmlstream_error(t, "You didn't send valid XML.");
        else if(ns == NULL)
            xmlstream_error(t, "You didn't specify a default xmlns='' namespace.");
        else if(to == NULL)
            xmlstream_error(t, "You didn't send your to='server' attribute.");
        else if(handler == NULL)
            xmlstream_error(t, "This server doesn't know about your namespace.");
        else if(!js_config_name(C_CHECK,to))
            xmlstream_error(t, "This server is not the one in your to attribute.");

    } else {

        /* let the real handler be the welcome party */
        (handler)(t, in, type);

        /* we have a valid handler, hand off processing of subsequent nodes */
        ts->data = (void *)handler;

    }

}

