/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 *
 *  svc_client.c -- service to handle connections from jabber clients
 *
 */

#include "jserver.h"
#include "svc_client_ssl.h"

/* data about a client connection */
typedef struct svc_client_data_struct
{
    session s;      /* the session */
    char *server;   /* host name the client connected to */
    char *id;       /* the user's jabber id */
    rlimit rate;    /* packet rate limit */
} *clidata, _clidata;


/*
 *  svc_client_onSend -- send a packet a packet to the client
 *  
 *  This function gets called by a worker thread to send packets
 *  back to the client.
 *
 *  parameters
 *      s -- the session to send the packet to
 *      p -- the packet to send
 *      arg -- the thread 
 *
 */
void svc_client_onSend(session s, jpacket p, void *arg)
{
    thread t_socket = (thread)arg;  /* cast the thread to correct type */
    tstream ts;                     /* the tstream to the client*/
    clidata cd;                     /* the client data */

    /* if there's no packet, don't try to send */
    if(p == NULL)
    {
        /* a null thread is a signal to exit from the session */
        if(arg != NULL)
        {
            /* debug message */
            log_debug(ZONE,"session exited externally");

            /* dig buried structures out */
            ts = (tstream)(t_socket->data);
            cd = (clidata)(ts->data_aux1);

            /* so that the cleanup handler knows the session is already gone, preventing race condition */
            cd->s = NULL;

            /* send an error to the client */
            xmlstream_error(t_socket,"forced disconnect");

        }
        return;
    }

    /* copy the from attribute from the packet to the xmlnode */
    if(p->from != NULL)
        xmlnode_put_attrib(p->x, "from", jid_full(p->from));

    /* hide the to/from if it's this resource */
    if(jid_cmp(p->to,s->id) == 0)
        xmlnode_hide_attrib(p->x, "to");
    if(p->type != JPACKET_MESSAGE && jid_cmp(p->from,s->id) == 0)
        xmlnode_hide_attrib(p->x, "from");

    /* log the packet */
    log_debug(ZONE,"SVC_BASIC: got a packet[%s] to send to the user[%d]",xmlnode2str(p->x),t_socket);

    /* write the packet to the client's session */
    xmlstream_write(t_socket, p->x);

}

/*
 *  svc_client_cleanup -- clean up the client session
 *
 *  This function gets called by pth when the session's thread gets killed,
 *  either on a disconnect from the client, or because of an error
 *
 *  parameters
 *      arg -- this is a pointer to the thread controlling the session
 *             it is treated as an opaque pointer by pth, so it needs
 *             to be cast back into a thread 
 *
 */
void svc_client_cleanup(void *arg)
{
    thread t = (thread)arg;                 /* pointer to the session thread */
    tstream ts = (tstream)(t->data);        /* extract the tstream from the thread */
    clidata cd = (clidata)(ts->data_aux1);  /* extract the client data from the stream */

    /* clean up SSL stuff if necessary */
#ifdef HAVE_OPENSSL

    /* the SSL data is stored in the tstream */
    SSL *ssl = (SSL *)(ts->data_aux4);

    /* clean up the SSL data, if necessary */
    if(ssl != NULL)
    {
        SSL_set_shutdown(ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
        SSL_free(ssl);
        ERR_remove_state(0);
    }

#endif

    /* if the session is still alive, shut it down */
    log_debug(ZONE,"SVC_BASIC: cleanup %X",cd);
    if(cd->s != NULL)
    {
        /* so that onSend knows the thread is gone, preventing race condition */
        cd->s->arg = NULL;

        /* clean up the session gracefully */
        js_session_end(cd->s,"disconnected");

    }

    rate_free(cd->rate);
}


/*
 *  svc_client_onNode -- handle node from jabber client
 *
 *  The connection/stream has passed all tests by the server when it get's here,
 *  so we know we're dealing with a valid XML stream. This is an xmlstream_onNode
 *  handler that gets called by js_conn_negotiate() when the <stream/> root node
 *  is recieved and again by the xmlstream for each subsequent node in the stream.
 *
 *  parameters
 *      t -- the thread for this connection
 *      in -- the xmlnode received from the client
 *      type -- the type of the noded, #defined in etherx.h
 *      
 */
void svc_client_onNode(thread t, xmlnode in, int type)
{

    tstream ts = (tstream)(t->data);        /* fetch the tstream from the thread */
    clidata cd = (clidata)(ts->data_aux1);  /* fetch the client data from the tstream */
    xmlnode query;                          /* */
    jpacket p;
    jid id;
    xmlnode x;
    sreturn sr;

    /* debug message */
    log_debug(ZONE,"SVC_BASIC: socket %d thread %X onNode %X",ts->fd,t->id,cd);

    /* is this the <stream/> node? */
    if(type == XMLSTREAM_ROOT)
    {
        /* register a function to clean up data structures when the connection closes */
        pool_cleanup(t->p, svc_client_cleanup, (void *)t);
        pool_label(t->p,"svc_client",0);

        /* initialize client data */
        cd = ts->data_aux1 = pmalloc(t->p, sizeof(_clidata));
        cd->s = NULL;
        cd->server = pstrdup(t->p,xmlnode_get_attrib(in, "to"));
        x = xmlstream_header(NS_CLIENT,NULL,xmlnode_get_attrib(in, "to"));
        cd->id = pstrdup(t->p,xmlnode_get_attrib(x,"id"));
        cd->rate = rate_new(60,30); /* 30 packets in a 60 second period */

        /* open the output stream back to the client */
        xmlstream_write_header(t, x);
        return;

    }

    /* was this the </stream> node? */
    if(type == XMLSTREAM_CLOSE)
    {
        /* the connection was closed */
        return;
    }

    /* is this data that didn't get written to the stream before it closed? */
    if(type == XMLSTREAM_BOUNCE)
    {
        /* XXX leftovers that didn't get sent */
        log_debug(ZONE,"leftovers: %s",xmlnode2str(in));
        xmlnode_free(in);
        return;
    }

    /* was there a parse error in the incoming XML? */
    if(type == XMLSTREAM_ERROR)
    {
        /* send an error message and close the connection */
        xmlstream_error(t, "You sent malformed XML");
        return;
    }

    /* if there is no session, the user hasn't logged in */
    if(cd->s == NULL)
    {
        /* stash the session id in the incomming node -- used by auth functions */
        xmlnode_put_attrib(in,"sid",cd->id);

        /* create a jserver packet from the node */
        p = jpacket_new(in);

        /* so we can access the query data later */
        query = p->iq;

        /*
         * try to authorize the user
         *
         * js_service_prescreen() doesn't have any knowledge of where
         * the auth request is coming from, so it can't send an appropriate
         * response. Instead, it modifies the xmlnode in the packet it's given
         * to be an appropriate response.
         */
        sr = js_service_prescreen(p);

        /* hide the session id so it doesn't get sent back to the client */
        xmlnode_hide_attrib(in,"sid");

        /* is the user authed? */
        switch(sr)
        {
            /* authentication failed */
        case S_BOUNCE:

            /* log a debug message */
            log_debug(ZONE,"prescreening returned data, bouncing back");

            /* send an error message back to the client */
            xmlstream_write(t,in);
            break;

            /* the user has been authenticated */
        case S_AUTHED:

            /* make a temp jid for this new user */
            id = jid_new(xmlnode_pool(in),js__hostname);
            jid_set(id,xmlnode_get_tag_data(query,"username"),JID_USER);
            jid_set(id,xmlnode_get_tag_data(query,"resource"),JID_RESOURCE);

            /* create a user session */
            log_debug(ZONE,"creating new session for %s",jid_full(id));
            cd->s = js_session_new(id, svc_client_onSend, (void *)t);

            /* send a result iq back to the client */
            xmlstream_write(t,in);
            break;
        default:
        }
        return;
    }

    /* check the rate limit */
    if(rate_check(cd->rate,"a",1)) pth_sleep(2); /* two second penalty */

    /* there are no special conditions, pass the packet on to the session to handle */
    log_debug(ZONE,"sending packet to session: %s",xmlnode2str(in));
    js_session_from(cd->s, jpacket_new(in));

}

/*
 *  svc_client -- initialize the service
 *  This function gets called in js_static as part of the server start up routine.
 *  It registers svc_client_onNode() as the node handler for the jabber:client namespace
 *
 */
void svc_client(void)
{
    log_debug(ZONE,"svc_client: init()");

    js_conn_namespace(C_SET,NS_CLIENT,(xmlstream_onNode)svc_client_onNode);
#ifdef HAVE_OPENSSL
    log_debug(ZONE, "svc_client: init() SSL functions");
#endif
}


