#include "svc_http.h"

/* HTTP session manager tools */

hs svc_http__sessions = NULL;

/* incoming data from the session */
void svc_http_session_in(session s, jpacket p, void *arg)
{
    hs h = (hs)arg;
    jpcache cur;
    hrq rq;

    /* server says the session ended (XXX improve) */
    if(p == NULL)
    {
        /* unlink from the main ring */
        h->prev->next = h->next;
        h->next->prev = h->prev;

        /* disconnect any RT connections */
        for(rq = h->rt; rq != NULL; rq = rq->next)
        {
            rq->s = NULL;
            tstream_close(rq->t);
        }

        /* free the caches, redeliver messages */
        for(cur = h->cache->next; cur != h->cache; cur = cur->next)
        {
            if(cur->p->type == JPACKET_MESSAGE)
                js_deliver(cur->p); /* redeliver unsent messages */
            else
                xmlnode_free(cur->p->x);
        }
        return;
    }

    log_debug(ZONE,"sent %s",xmlnode2str(p->x));

    /* testing RT */
    for(rq = h->rt; rq != NULL; rq = rq->next)
    {
        log_debug(ZONE,"RT close %X",rq);
        tstream_close(rq->t);
    }

    /* store in the cache */
    svc_http_cache(h,p);
}

/* create a new http session with this /session request */
void _svc_http_session_new(hrq rq)
{
    char *user, *pass, *res;
    hs h;
    char str[64];
    jpacket p;
    session s;
    jid id;

    /* get auth from query string or basic auth */
    user = xmlnode_get_attrib(rq->q,"user");
    if(user == NULL)
        user = xmlnode_get_attrib(rq->x,"username");
    pass = xmlnode_get_attrib(rq->q,"pass");
    if(pass == NULL)
        pass = xmlnode_get_attrib(rq->x,"password");

    /* get preferred or default resource */
    res = xmlnode_get_attrib(rq->q,"res");
    if(res == NULL)
        res = pstrdup(rq->p,"web");

    /* jserver auth */
    p = jpacket_new(jutil_iqnew(JPACKET__SET,NS_AUTH));
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "username"),user,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "password"),pass,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "resource"),res,-1);
    log_debug(ZONE,"attempting to create session with auth %s",xmlnode2str(p->x));

    /* setup reply body */
    rq->body = xmlnode_new_tag_pool(rq->p,"jabber");

    if(js_service_prescreen(p) != S_AUTHED)
    { /* loozer */
        xmlnode_put_attrib(rq->body,"error",xmlnode_get_tag_data(p->x,"error"));
        xmlnode_free(p->x);
        return;
    }

    /* create a temp jid and start the session */
    id = jid_new(p->p,js__hostname);
    jid_set(id,user,JID_USER);
    jid_set(id,res,JID_RESOURCE);
    s = js_session_new(id,svc_http_session_in,NULL);
    xmlnode_free(p->x);

    /* create http session holder */
    h = pmalloc(s->p,sizeof(_hs));
    memset(h,0,sizeof(_hs));
    h->s = s;
    h->timeout = HTTP_MINWAIT;
    h->last = time(NULL);
    s->arg = (void *)h; /* like, we need this :) */
    rq->s = h;

    /* create a random sid */
    sprintf(str,"%d",rand());
    sprintf(str,"%s",shahash(str));
    h->sid = pstrdup(s->p,str);
    xmlnode_put_attrib(rq->body,"sid",h->sid);

    /* store user/pass for 401 auth */
    h->user = pstrdup(s->p,user);
    h->pass = pstrdup(s->p,pass);

    /* insert into timeout ring */
    h->next = svc_http__sessions->next;
    svc_http__sessions->next->prev = h;
    h->prev = svc_http__sessions;
    svc_http__sessions->next = h;

    /* create cache stub */
    h->cache = pmalloc(s->p,sizeof(_jpcache));
    h->cache->next = h->cache->prev = h->cache;
    h->cache->p = NULL;

    /* optional initial presence */
    if(xmlnode_get_attrib(rq->q,"status") != NULL)
        js_session_from(s,jpacket_new(jutil_presnew(JPACKET__AVAILABLE,NULL,xmlnode_get_attrib(rq->q,"status"))));

    /* optional initial roster fetch */
    if(xmlnode_get_attrib(rq->q,"roster") != NULL)
        js_session_from(s,jpacket_new(jutil_iqnew(JPACKET__GET,NS_ROSTER)));

}

/* dumb wrapper that replies */
void svc_http_session_new(hrq rq)
{
    _svc_http_session_new(rq);
    svc_http_session_reply(rq);
}

void svc_http_session_end(hs h)
{
}

/* try with sid */
void _svc_http_session_sid(hrq rq)
{
    char *sid;
    hs cur;

    /* first attempt to find it based on the sid */
    sid = xmlnode_get_attrib(rq->q,"sid");
    for(cur = svc_http__sessions->next;cur != svc_http__sessions;cur = cur->next)
    {
        if(j_strcmp(sid,cur->sid) == 0)
        {
            rq->s = cur;
            return;
        }
    }
}

/* lookup on basic auth user/pass */
void _svc_http_session_basic(hrq rq)
{
    hs cur;
    char *user, *pass;

    user = xmlnode_get_attrib(rq->x,"username");
    pass = xmlnode_get_attrib(rq->x,"password");

    if(user == NULL || pass == NULL) return;

    for(cur = svc_http__sessions->next;cur != svc_http__sessions;cur = cur->next)
    {
        if(j_strcmp(user,cur->user) == 0 && j_strcmp(pass,cur->pass) == 0)
        {
            rq->s = cur;
            return;
        }
    }

    /* start a session on the fly if we can */
    _svc_http_session_new(rq);
}

/* find the right session and pass on to the client handlers */
void svc_http_session(hrq rq)
{
    /* try sid first */
    if(xmlnode_get_attrib(rq->q,"sid") != NULL)
        _svc_http_session_sid(rq);

    /* try basic auth */
    if(rq->s == NULL)
        _svc_http_session_basic(rq);

    if(rq->s == NULL)
    {
        rq->body = xmlnode_new_tag_pool(rq->p,"jabber");
        xmlnode_put_attrib(rq->body,"error","Session Timed Out");
        return;
    }

    log_debug(ZONE,"located session %s",jid_full(rq->s->s->id));

    /* reset timout */
    rq->s->last = time(NULL);

    /* setup reply body */
    rq->body = xmlnode_new_tag_pool(rq->p,"jabber");
    xmlnode_put_attrib(rq->body,"sid",rq->s->sid);
}

/* poll and timeout inactive sessions */
void *svc_http_session_main(void *arg)
{
    hs cur;
    time_t t;

    svc_http__sessions = malloc(sizeof(_hs)); /* forever */
    svc_http__sessions->next = svc_http__sessions->prev = svc_http__sessions; /* init ring */

    while(1)
    {
        pth_sleep(HTTP_MINWAIT);
        t = time(NULL);
        for(cur = svc_http__sessions->next;cur != svc_http__sessions;cur = cur->next)
        {
            if(cur->rt == NULL && (t - cur->last) > cur->timeout)
            {
                log_debug(ZONE,"Timing out stale session %s",jid_full(cur->s->id));
                svc_http_session_end(cur);
                js_session_end(cur->s,"Timed Out");
            }
        }
    }
}

void svc_http_session_init(void)
{
    pth_attr_t attr;

    /* for better session ids */
    srand(time(NULL));

    /* start the timeout thread */
    attr = pth_attr_new();
    pth_attr_set(attr, PTH_ATTR_JOINABLE, FALSE);
    pth_attr_set(attr, PTH_ATTR_STACK_SIZE, 16*1024);
    pth_spawn(attr, svc_http_session_main, NULL); /* times out sessions */
    pth_attr_destroy(attr);
}


/* escape the damn javascript strings! ugh */
char *_apos(pool p, char *buf)
{
    int i,j,oldlen,newlen;
    char *temp;

    if(p == NULL || buf == NULL) return(NULL);

    oldlen = newlen = strlen(buf);
    for(i=0;i<oldlen;i++)
        if(buf[i] == '\'')
            newlen += 2;

    if(oldlen == newlen) return buf;

    temp = pmalloc(p,newlen+1);

    if(temp==NULL) return(NULL);

    for(i=j=0;i<oldlen;i++)
        if(buf[i] == '\'')
        {
            memcpy(&temp[j],"\\'",2);
            j += 2;
        }else{
            temp[j++] = buf[i];
        }

    temp[j] = '\0';
    return temp;
}

/* preprocess the reply, converting the xml body into other formats, used for session stuff primarily */
void svc_http_session_reply(hrq rq)
{
    xmlnode cur, cur2;
    int i = 0;
    char strid[32], *pass, *sid;
    spool sp;

    if(rq->body == NULL) return svc_http_reply(rq);

    if(rq->s != NULL)
        sid = rq->s->sid;
    else
        sid = NULL;

    switch(rq->ext[0])
    {
    case 'x': /* send the xml */
        svc_http_reply_mime(rq,MIME_XML);
        svc_http_reply_append(rq,xmlnode2str(rq->body),rq);
        break; /* convert to HTML */
    case 'h':
        svc_http_reply_mime(rq,MIME_HTML);
        svc_http_reply_append(rq,"<HTML><BODY>",rq);
        if(sid != NULL)
            svc_http_reply_append(rq,"<H1>Session: ",sid,"</H1>",rq);
        else
            svc_http_reply_append(rq,"<H1>Error: ",xmlnode_get_attrib(rq->body,"error"),"</H1>",rq);
        for(cur = xmlnode_get_firstchild(rq->body); cur != NULL; cur = xmlnode_get_nextsibling(cur))
        {
            sprintf(strid,"jabber[%d]",i++);
            svc_http_reply_append(rq,"<H2>",xmlnode_get_name(cur),rq);
            if(xmlnode_get_attrib(cur,"type") != NULL)
                svc_http_reply_append(rq,"<SUP>(",xmlnode_get_attrib(cur,"type"),")</SUP>",rq);
            svc_http_reply_append(rq,"</H2>From: ",xmlnode_get_attrib(cur,"from"),"<BR>",rq);
            switch(*xmlnode_get_name(cur))
            {
            case 'm':
                if(xmlnode_get_tag(cur,"subject") != NULL)
                    svc_http_reply_append(rq,"Subject: ",xmlnode_get_tag_data(cur,"subject"),"<BR>",rq);
                if(xmlnode_get_tag(cur,"body") != NULL)
                    svc_http_reply_append(rq,"<BLOCKQUOTE>",xmlnode_get_tag_data(cur,"body"),"</BLOCKQUOTE>",rq);
                break;
            case 'p':
                if(xmlnode_get_tag(cur,"status") != NULL)
                    svc_http_reply_append(rq,"Status: ",xmlnode_get_tag_data(cur,"status"),"<BR>",rq);
                if(xmlnode_get_tag(cur,"show") != NULL)
                    svc_http_reply_append(rq,"Show: ",xmlnode_get_tag_data(cur,"show"),"<BR>",rq);
                break;
            case 'i':
                if(xmlnode_get_attrib(cur,"ask") != NULL)
                    svc_http_reply_append(rq,"Asking: ",xmlnode_get_attrib(cur,"ask"),"<BR>",rq);
                for(cur2 = xmlnode_get_firstchild(cur); cur != NULL; cur = xmlnode_get_nextsibling(cur))
                    if(xmlnode_get_type(cur) == NTYPE_TAG)
                        svc_http_reply_append(rq,"Group: ",xmlnode_get_data(cur),"<BR>",rq);
                break;
            }
        }
        svc_http_reply_append(rq,"</BODY></HTML>",rq);
        break;
    case 'j': /* convert to javascript */
        svc_http_reply_mime(rq,MIME_JS);
        svc_http_reply_append(rq,"jabber = new Array(); jabber.sid = '",sid,"'; jabber.error = '",xmlnode_get_attrib(rq->body,"error"),"'; ",rq);
        for(cur = xmlnode_get_firstchild(rq->body); cur != NULL; cur = xmlnode_get_nextsibling(cur))
        {
            sprintf(strid,"jabber[%d]",i++);
            svc_http_reply_append(rq,strid," = new Array(); ",strid,".packet = '",xmlnode_get_name(cur),"'; ",rq);
            svc_http_reply_append(rq,strid,".type = '",xmlnode_get_attrib(cur,"type"),"'; ",rq);
            svc_http_reply_append(rq,strid,".from = '",_apos(rq->p,xmlnode_get_attrib(cur,"from")),"'; ",rq);
            switch(*xmlnode_get_name(cur))
            {
            case 'm':
                svc_http_reply_append(rq,strid,".subject = '",_apos(rq->p,xmlnode_get_tag_data(cur,"subject")),"'; ",rq);
                svc_http_reply_append(rq,strid,".body = '",_apos(rq->p,xmlnode_get_tag_data(cur,"body")),"'; ",rq);
                break;
            case 'p':
                svc_http_reply_append(rq,strid,".status = '",_apos(rq->p,xmlnode_get_tag_data(cur,"status")),"'; ",rq);
                svc_http_reply_append(rq,strid,".show = '",_apos(rq->p,xmlnode_get_tag_data(cur,"show")),"'; ",rq);
                break;
            case 'i':
                svc_http_reply_append(rq,strid,".ask = '",rq->p,xmlnode_get_attrib(cur,"ask"),"'; ",rq);
                svc_http_reply_append(rq,strid,".name = '",rq->p,xmlnode_get_attrib(cur,"name"),"'; ",rq);
                svc_http_reply_append(rq,strid,".group = '",_apos(rq->p,xmlnode_get_tag_data(cur,"group")),"'; ",rq);
                break;
            }
        }
        break;
    case 'n': /* null, no response,  */
        svc_http_reply_mime(rq,MIME_HTML);
        svc_http_reply_append(rq,"&nbsp;",rq);
        break;
    case 'c': /* simple redirect */
        if(xmlnode_get_attrib(rq->body,"error") != NULL)
            return svc_http_redirect(rq,xmlnode_get_attrib(rq->q,"fail"));
        pass = xmlnode_get_attrib(rq->q,"pass");
        /* if they pass a query string, append the sid */
        if(sid != NULL && strchr(pass,'?') != NULL)
        {
            sp = spool_new(rq->p);
            spooler(sp,pass,sid,sp);
            xmlnode_put_attrib(rq->q,"pass",spool_print(sp));
        }
        return svc_http_redirect(rq,xmlnode_get_attrib(rq->q,"pass"));
        break;
    }

    svc_http_reply(rq);
}

