#include "svc_irc.h"

/* OUT: IRC --> Jabber XML */

/* start a new session for this connection */
void svc_irc_session(ircdata d)
{
    /* set up the jabber id and start the session */
    d->id = jid_new(d->t->p,d->server);
    jid_set(d->id, d->nick, JID_USER);
    jid_set(d->id, d->res, JID_RESOURCE);
    d->s = js_session_new(d->id,svc_irc_in,(void *)d);

    /* invite them to the main room */
    svc_irc_line(d,NULL,"INVITE",ROOM,1,d->nick);
    d->phase = PHASE_PRE;
}

void svc_irc_help_room(ircdata d)
{
    svc_irc_line(d,"nickserv","NOTICE","-------------------------------------------------------",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Your roster is the users in this room, which you can manage with:",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","    /msg nickserv add|del|set user@server [nick]",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","You can manage your presence subscriptions with:",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","    /msg nickserv [un]subscribe[d] user@server [message]",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","You can register with agents by typing:",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","    /msg agents get",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","And follow the instructions you receive",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Send messages to users using 'nick: message' or '/msg nick message'",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Change presence by '/me on|off|away|xa|dnd status message' (you are currently 'on')",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","-------------------------------------------------------",1,d->nick);
}

/* create the room */
void svc_irc_room(ircdata d)
{
    jpacket p;

    /* send the new room info */
    svc_irc_line(d,d->nick,"JOIN",ROOM,0);
    svc_irc_line(d,NULL,"332",jid_full(d->id),2,d->nick,ROOM); /* title */

    /* get the roster */
    p = jpacket_new(jutil_iqnew(JPACKET__GET,NS_ROSTER));
    js_session_from(d->s,p);

    /* set the presence */
    p = jpacket_new(xmlnode_new_tag("presence"));
    js_session_from(d->s,p);

    svc_irc_line(d,NULL,"NOTICE","Welcome!",1,d->nick);
    svc_irc_help_room(d);

    d->phase = PHASE_NORM;
}

/* process a login, /msg login password [resource] */
int svc_irc_auth(ircdata d)
{
    jpacket p;

    p = jpacket_new(jutil_iqnew(JPACKET__SET,NS_AUTH));
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "username"),d->nick,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "password"),d->pass,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "resource"),d->res,-1);

    log_debug(ZONE,"auth %s",xmlnode2str(p->x));

    if(js_service_prescreen(p) == S_AUTHED)
    { /* start the session */
        xmlnode_free(p->x);
        return 1;
    }

    xmlnode_free(p->x);
    return 0;
}

/* process a registration, /msg register password  */
int svc_irc_reg(ircdata d)
{
    int ret;
    jpacket p;

    p = jpacket_new(jutil_iqnew(JPACKET__SET,NS_REGISTER));
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "username"),d->nick,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(p->iq, "password"),d->pass,-1);

    log_debug(ZONE,"register %s",xmlnode2str(p->x));
    js_service_prescreen(p);

    jpacket_reset(p);
    ret = jpacket_subtype(p);
    xmlnode_free(p->x);

    if(ret == JPACKET__RESULT)
        return 1;
    else
        return 0;
}

/* process session presence bit */
void svc_irc_pres(ircdata d, char *data)
{
    char *desc;
    xmlnode x, item;
    jid id;

    desc = strstr(data," ");
    if(desc != NULL)
    {
        *desc = '\0';
        ++desc;
    }

    if(j_strcmp(data,"off") == 0)
    {
        x = jutil_presnew(JPACKET__UNAVAILABLE,NULL,desc);

        /* whack the presence cache */
        for(item = xmlnode_get_firstchild(d->roster); item != NULL; item = xmlnode_get_nextsibling(item))
        { /* make sure all the users are unavail */
            if(xmlnode_get_attrib(item,"jid") == NULL) continue;

            id = jid_new(xmlnode_pool(x),xmlnode_get_attrib(item,"jid"));
            if(ppdb_primary(d->pres,id) != NULL)
                svc_irc_line(d, "presence", "MODE", NULL, 3, ROOM, "-o", xmlnode_get_attrib(item,"name"));
        }
        ppdb_free(d->pres);
        d->pres = NULL;
    }else{
        x = jutil_presnew(JPACKET__AVAILABLE,NULL,desc);
        if(j_strcmp(data,"on") != 0)
            xmlnode_insert_cdata(xmlnode_insert_tag(x,"show"),data,-1);
    }

    js_session_from(d->s,jpacket_new(x));
}

/* process a PRIVMSG */
void svc_irc_message(ircdata d, char *to, char *data)
{
    xmlnode x, body;
    jpacket p;
    int chat = 1;
    char *colon, *space;

    log_debug(ZONE,"message to %s data %X %s",to, *data, data);

    if(j_strcmp(to,ROOM) == 0)
    {
        colon = strstr(data,":");
        space = strstr(data," ");
        if(colon == NULL || (space != NULL && space < colon))
        {
            if(d->lastnick == NULL)
            {
                svc_irc_line(d,NULL,"NOTICE","You need to address someone specifically with nick:",1,d->nick);
                return;
            }
            to = d->lastnick;
        }else{
            *colon = '\0';
            colon++;
            to = data;
            data = colon;
            free(d->lastnick);
            d->lastnick = strdup(to);
        }
        chat = 0;
    }

    if(svc_irc_id(d,to) != NULL)
        to = svc_irc_id(d,to);

    x = xmlnode_new_tag("message");
    xmlnode_put_attrib(x, "to", to);
    if(chat)
        xmlnode_put_attrib(x,"type","chat");
    if(data != NULL)
    {
        body = xmlnode_insert_tag(x, "body");
        xmlnode_insert_cdata(body,data,strlen(data));
    }

    log_debug(ZONE,"sending %s",xmlnode2str(x));

    p = jpacket_new(x);
    js_session_from(d->s,p);
}

/* process roster requests, /msg nickserv add|del|set user@server [nick] */
void svc_irc_roster(ircdata d, char *data)
{
    char *ids, *nick;
    xmlnode item;
    jpacket p;
    jid id;

    ids = strstr(data," ");
    if(ids == NULL) return;
    *ids = '\0';
    ids++;

    nick = strstr(ids," ");
    if(nick != NULL)
    {
        *nick = '\0';
        nick++;
    }

    p = jpacket_new(jutil_iqnew(JPACKET__SET,NS_ROSTER));
    id = jid_new(p->p,ids);
    item = svc_irc_item(d,id);

    if(item == NULL)
    {
        item = xmlnode_insert_tag(p->iq,"item");
        xmlnode_put_attrib(item,"jid",ids);
    }else{
        item = xmlnode_insert_tag_node(p->iq,item);
    }

    if(nick != NULL)
        xmlnode_put_attrib(item,"name",nick);

    if(j_strcmp(data,"del") == 0)
        xmlnode_put_attrib(item,"subscription","remove");

    js_session_from(d->s,p);
}

/* process presence subscription requests, /msg nickserv [un]subscribe[d] user@server description... */
void svc_irc_s10n(ircdata d, char *data)
{
    char *id, *desc;
    xmlnode x;

    id = strstr(data," ");
    if(id == NULL) return;
    *id = '\0';
    id++;

    desc = strstr(id," ");
    if(desc != NULL)
    {
        *desc = '\0';
        desc++;
    }

    x = xmlnode_new_tag("presence");
    xmlnode_put_attrib(x, "to", id);
    xmlnode_put_attrib(x, "type", data);
    if(desc != NULL)
        xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),desc,strlen(desc));

    js_session_from(d->s,jpacket_new(x));
}

/* handle who queries */
void svc_irc_who(ircdata d, char *to)
{
    xmlnode cur;
    char *ids, *nick, *user, *server;
    pool p;
    jid id;

    log_debug(ZONE,"handling who query to %s",to);

    /* /who ROOM */
    if(j_strcmp(to,ROOM) == 0)
    {
        if(d->roster == NULL)
        {
            d->whoflag = 1;
            return;
        }
        p = pool_new();
        for(cur = xmlnode_get_firstchild(d->roster); cur != NULL; cur = xmlnode_get_nextsibling(cur))
        {
            if((ids = xmlnode_get_attrib(cur,"jid")) != NULL)
            {
                id = jid_new(p,ids);
                nick = svc_irc_nick(d, id);
                user = strstr(nick,"!");
                *user = '\0';
                ++user;
                server = strstr(user,"@");
                *server = '\0';
                ++server;
                svc_irc_line(d,NULL,"352",NULL,9,d->nick,ROOM,user,server,d->server,nick,"H",":0",ids);
            }
        }
        svc_irc_line(d,NULL,"315","End of /who list",2,d->nick,ROOM);
        pool_free(p);

        return;
    }

    /* /who user */

}

void svc_irc_help_auth(ircdata d)
{
    svc_irc_line(d,"nickserv","NOTICE","-------------------------------------------------------",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","You need to have an account to utilize this service.",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Create a new account using your current nickname:",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","    /msg nickserv register password",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","Or authenticate your nickname with an existing account:",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","    /msg nickserv identify password [resource]",1,d->nick);
    svc_irc_line(d,"nickserv","NOTICE","-------------------------------------------------------",1,d->nick);
}
void svc_irc_welcome(ircdata d)
{
    svc_irc_line(d,NULL,"001","Welcome to Jabber's IRC Service",1,d->nick);
    svc_irc_line(d,NULL,"004",NULL,5,d->nick,xmlnode_get_data(js_config("names/default")),"1.2.3","foo","bar");
}

void svc_irc_agents(ircdata d,char *data)
{
    if(data==NULL) return;
    if(d->agent_fields==NULL||d->next_field==NULL)
    { /* this is a new agent request */
        xmlnode iq;
        jpacket p;
        iq=xmlnode_new_tag("iq");
        xmlnode_put_attrib(iq,"to",data);
        xmlnode_put_attrib(iq,"type","get");
        xmlnode_put_attrib(xmlnode_insert_tag(iq,"query"),"xmlns","jabber:iq:register");
        p=jpacket_new(iq);
        js_session_from(d->s,p);
        svc_irc_line(d,"agents","NOTICE","Sent a request to that agent.",1,d->nick);
        return;
    }
    /* else, this is data for the next field */
    if(j_strcmp(data,"remove")==0)
    { /* send a <remove/> flag */
        xmlnode iq;
        jpacket p;
        iq=xmlnode_new_tag("iq");
        xmlnode_put_attrib(iq,"to",d->agent_jid);
        xmlnode_put_attrib(iq,"type","set");
        xmlnode_put_attrib(xmlnode_insert_tag(iq,"query"),"xmlns","jabber:iq:register");
        xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(iq,"query"),"key"),d->key,-1);
        xmlnode_insert_tag(xmlnode_get_tag(iq,"query"),"remove");
        p=jpacket_new(iq);
        js_session_from(d->s,p);
        svc_irc_line(d,"agents","NOTICE","Sent a Remove request to the agent.",1,d->nick);
        d->agent_phase=0;
        if(d->agent_fields!=NULL&&d->agent_fields->p!=NULL)
        {
            pool_free(d->agent_fields->p);
            d->next_field=NULL;
            d->agent_fields=NULL;
            d->agent_jid=NULL;
            d->key=NULL;
        }
        return;
    }
    else if(j_strcmp(data,"cancel")==0)
    {
        d->agent_phase=0;
        if(d->agent_fields!=NULL&&d->agent_fields->p!=NULL)
        {
            pool_free(d->agent_fields->p);
            d->next_field=NULL;
            d->agent_fields=NULL;
            d->agent_jid=NULL;
            d->key=NULL;
        }
        svc_irc_line(d,"agents","NOTICE","Request Canceled",1,d->nick);
        return;
    }
    else if(strncasecmp(data,"back",4)==0)
    {
        if(d->next_field==NULL)
            return;
        if(d->next_field->prev!=NULL)
            d->next_field=d->next_field->prev;
        else
            svc_irc_line(d,"agents","NOTICE","You are at the start",1,d->nick);
    }
    else if(strncasecmp(data,"skip",4)==0)
    {
        if(d->next_field==NULL)
            return;
        if(d->next_field->next!=NULL)
            d->next_field=d->next_field->next;
        else
        {
            d->next_field=d->next_field->next;
            svc_irc_line(d,"agents","NOTICE","You are at the end, sending in your form.",1,d->nick);
        }
    }
    else
    {
        d->next_field->data=pstrdup(d->next_field->p,data);
        d->next_field=d->next_field->next;
    }
    /* if this is the end, send in the registration */
    if(d->next_field==NULL)
    {
        jpacket p;
        xmlnode iq,query;
        iq=xmlnode_new_tag("iq");
        query=xmlnode_insert_tag(iq,"query");
        xmlnode_put_attrib(query,"xmlns","jabber:iq:register");
        xmlnode_put_attrib(iq,"to",d->agent_jid);
        xmlnode_put_attrib(iq,"type","set");
        d->next_field=d->agent_fields;
        svc_irc_line(d,"agents","NOTICE","Registration complete, sending in this form:",1,d->nick);
        while(d->next_field!=NULL)
        {
            spool sp=spool_new(d->next_field->p);
            spooler(sp,"Field: ",d->next_field->name,", Data: ",d->next_field->data,sp);
            svc_irc_line(d,"agents","PRIVMSG",spool_print(sp),1,d->nick);
            xmlnode_insert_cdata(xmlnode_insert_tag(query,d->next_field->name),d->next_field->data,-1);
            d->next_field=d->next_field->next;
        }
        xmlnode_insert_cdata(xmlnode_insert_tag(query,"key"),d->key,-1);
        p=jpacket_new(iq);
        js_session_from(d->s,p);
        svc_irc_line(d,"agents","NOTICE","Registration set. To send another, type 'get'",1,d->nick);
    }
    else
    { /* display the next line */
        spool sp=spool_new(d->next_field->p);
        svc_irc_line(d,"agents","NOTICE" ,"-------------------------------------------------------",1,d->nick);
        svc_irc_line(d,"agents","PRIVMSG","Type 'back' for the previous item",1,d->nick);
        svc_irc_line(d,"agents","PRIVMSG","Type 'skip' to skip this item",1,d->nick);
        svc_irc_line(d,"agents","PRIVMSG","Type 'cancel' to kill this request, or",1,d->nick);
        svc_irc_line(d,"agents","PRIVMSG","Type 'remove' to remove an existing registration",1,d->nick);
        svc_irc_line(d,"agents","NOTICE" ,"-------------------------------------------------------",1,d->nick);
        spooler(sp,"Current Item --=>",d->next_field->name,"<=-- Current Data: ",d->next_field->data,sp);
        svc_irc_line(d,"agents","PRIVMSG",spool_print(sp),1,d->nick);
        svc_irc_line(d,"agents","NOTICE" ,"-------------------------------------------------------",1,d->nick);
        svc_irc_line(d,"agents","NOTICE","Enter the new data for this item:",1,d->nick);
    }
}


void svc_irc_out(ircdata d, char *cmd, char *from, char **args)
{
    char *space, *space2;

    log_debug(ZONE,"command %s to %s with args[0] %s",cmd,from,args[0]);

    if(j_strcmp(cmd,"PONG") == 0)
        return;

    if(d->phase == PHASE_INIT)
    {
        if(j_strcmp(cmd,"USER") == 0 && args[0] != NULL)
        {
            /*
            args[0] username
                 1  hostname
                 2  servername
                 3  real name
            */
            if(d->nick==NULL)d->nick=pstrdup(d->t->p,args[0]);
            svc_irc_welcome(d);
            svc_irc_help_auth(d);
            d->phase = PHASE_LOGIN;
        }
        if(j_strcmp(cmd,"NICK")==0)
        { /* for those clients that send NICK first */
            d->nick=pstrdup(d->t->p,args[0]);
            svc_irc_line(d,d->nick,"NICK",args[0],0);
        }
        if(j_strcmp(cmd,"PASS")==0)
        { /* to support PASS functionality */
            if(d->nick!=NULL)
            { /* invalid after a USER or NICK command */
                svc_irc_line(d,"NickServ","462","Cannot Re-register",0);
            }
            else
            {
                d->pass=pstrdup(d->t->p,args[0]);
            }
        }
        if(d->nick!=NULL&&d->pass!=NULL)
        { /* this user supplied a pass and nick at connection */
            d->res=pstrdup(d->t->p,"irc");
            if(svc_irc_auth(d))
            {
                svc_irc_line(d,"NickServ","NOTICE","You Are Authorized on this server",1,d->nick);
                d->phase=PHASE_PRE;
                svc_irc_session(d);
            }
        }
        return;
    }

    if(d->phase == PHASE_LOGIN)
    {
        if(j_strcmp(cmd,"NICK") == 0)
        {
            /* hrmm....
            if(strlen(argv[0]) == 0) from = data;
            */
            svc_irc_line(d,d->nick,"NICK",args[0],0);
            d->nick = pstrdup(d->t->p, args[0]);
            return;
        }
        if(j_strcmp(cmd,"PRIVMSG") == 0 && args[0] != NULL && strcasecmp(args[0],"nickserv") == 0 && args[1] != NULL)
        {
            space = strstr(args[1]," ");
            if(space == NULL)
            {
                svc_irc_line(d,"nickserv","NOTICE","Failed!",1,d->nick);
                return;
            }
            *space = '\0';
            ++space;
            space2 = strstr(space," ");
            if(space2 == NULL)
            {
                d->res = pstrdup(d->t->p,"irc");
            }else{
                *space2 = '\0';
                ++space2;
                d->res = pstrdup(d->t->p,space2);
            }
            d->pass = pstrdup(d->t->p, space);

            if(strcasecmp(args[1],"identify") == 0 && svc_irc_auth(d))
            {
            }else if(strcasecmp(args[1], "register") == 0 && svc_irc_reg(d) && svc_irc_auth(d)){
                svc_irc_line(d,"nickserv","NOTICE","Account Created",1,d->nick);
            }else{
                svc_irc_line(d,"nickserv","NOTICE","Failed!",1,d->nick);
                return;
            }
            svc_irc_session(d);
        }else{
            svc_irc_help_auth(d);
        }
        return;
    }

    /* before you're in the room */
    if(d->phase == PHASE_PRE)
    {
        if(j_strcmp(cmd,"JOIN") == 0 && j_strcmp(args[0], ROOM) == 0)
            svc_irc_room(d);
        else
            svc_irc_line(d,NULL,"NOTICE","You may only /join and /msg at this point.",1,d->nick);
        return;
    }

    /* deal w/ normal traffic here */

    if(j_strcmp(cmd,"PRIVMSG") == 0 && args[0] != NULL && strcasecmp(args[0],"nickserv") == 0 && args[1] != NULL)
    { /* /msg nickserv crap */
        if(strstr(args[1],"subscribe") != NULL)
            svc_irc_s10n(d,args[1]);
        else
            svc_irc_roster(d,args[1]);
        return;
    }

    if(j_strcmp(cmd,"PRIVMSG") == 0 && args[0] != NULL &&  strcasecmp(args[0],"debug") == 0)
    { /* turn on debugging :) */
        d->debug ^= 1;
        return;
    }

    if(j_strcmp(cmd,"PRIVMSG")==0&&args[0]!=NULL&&strcasecmp(args[0],"agents")==0&&args[1]!=NULL)
    {

        if(j_strcmp(args[1],"get")==0)
        {
            jpacket p = jpacket_new(jutil_iqnew(JPACKET__GET,NS_AGENTS));
            js_session_from(d->s,p);
            return;
        }
        else
            svc_irc_agents(d,args[1]);
        return;
    }

    if(j_strcmp(cmd,"PRIVMSG") == 0 && *args[1] == 1)
    { /* CTCP stuff */
        /* cut off the CTCP bits */
        args[1][strlen(args[1]) - 1] = '\0';
        args[1]++;

        if(strncmp(args[1],"ACTION ",7) == 0)
        {
            if(j_strcmp(args[0],ROOM) == 0)
            { /* /me in the room controls presence */
                svc_irc_pres(d,args[1] + 7);
                return;
            }else{ /* make it a literal '/me' */
                *args[1] += 3;
                args[1][0] = '/';
                args[1][1] = 'm';
                args[1][2] = 'e';
            }
        }else{
            svc_irc_line(d,NULL,"NOTICE",NULL,3,d->nick,":I don't understand CTCP",args[1]);
            return;
        }
    }

    if(j_strcmp(cmd,"PRIVMSG") == 0 && args[0] != NULL)
    {
        svc_irc_message(d,args[0],args[1]);
        return;
    }

    if(j_strcmp(cmd,"WHO") == 0 && args[0] != NULL)
    {
        svc_irc_who(d,args[0]);
        return;
    }

    if(j_strcmp(cmd,"MODE") == 0)
        return;

    svc_irc_line(d,NULL,"NOTICE","I don't understand!",1,d->nick);
}

