/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 *
 *  xdb.c -- the user database API
 *
 *  To any xdb users, xdb is an xmlnode consumer. Any data coming from the xdb
 *  is managed by the xdb, any data sent back to it (which can be the same
 *  data/pointers only modified) is also then managed by xdb.
 * 
 *  Typically, you get the data with xdb_get, make your changes, then call xdb_set
 *  or just forget about it and move on.
 * 
 *  Notes for xdb_* components 
 *  First come first serve. All data returned in a get is aggressively cached,
 *  which means the get will not be called again, only the set when it changes.
 *  To prevent the caching, always make sure the (xdb)->cache bit is 1 (even in a set)
 *
 */

#include "jserver.h"

/*
 *  js_xdb_fetch -- find the xdb node for a particular xml namespace
 *  
 *  parameters
 *  	xdb -- the head of the xdb list
 *      ns -- the namespace to search for
 *
 *  returns
 *      the xdb node, if found 
 *      NULL if not
 */
xdb js_xdb_fetch(xdb cur, char *ns)
{
    /* for each node... */
    for(;cur != NULL; cur = cur->next)

        /* compare the namespaces */
        if(j_strcmp(cur->ns, ns) == 0)

            /* match! return this node */
            return cur;

    /* no matching namespace, return NULL */
    return NULL;

}

/*
 *  js_xdb -- manipulate user data
 *  
 *  Central switchboard. All commands go through here
 * 
 *  FIXME: The API is in transition... there will be
 *  changes after the 1.0 release.
 *
 */
xmlnode js_xdb(command cmd, udata user, char *ns, xmlnode data)
{
    static xlist l = NULL;
    static xdb global = NULL;
    xlist newl, i;
    xdb cur;

    /* initialize the user list */
    if(cmd == C_INIT)
    {
        log_debug(ZONE,"registered new xdb");
        newl = malloc(sizeof(_xlist));
        newl->c = (xcall)user;
        newl->arg = (void *)ns;
        newl->next = l;
        l = newl;
        return NULL;
    }

    /* first get the entry from the current cache, or create a cache entry */
    if(user != NULL)
    {
        cur = js_xdb_fetch(user->x_cache, ns);
        if(cur == NULL)
        {
            cur = pmalloc(user->p,sizeof(_xdb));
            cur->data = data;
            cur->cache = 1;
            cur->ns = pstrdup(user->p,ns);
            cur->next = user->x_cache;
            user->x_cache = cur;
        }
    }else{
        cur = js_xdb_fetch(global, ns);
        if(cur == NULL)
        {
            cur = malloc(sizeof(_xdb));
            cur->data = data;
            cur->cache = 1;
            cur->ns = strdup(ns);
            cur->next = global;
            global = cur;
        }
    }

    /* retreive user data from the db */
    if(cmd == C_GET)
    {
        log_debug(ZONE,"GET %s",ns);

        /* check cache first */
        if(cur->cache && cur->data != NULL)
        {
            /* cache data is valid */
            return cur->data;

        }else{

            /* free the stale data in the cache */
            xmlnode_free(cur->data);
            cur->data = NULL;

        }

        /* ask each xdb to search for the data until it's found */
        for(i = l; i != NULL; i = i->next)
            if((*(i->c))(0,cur,user,i->arg))
                return cur->data;

        /* cmd must be C_SET */
    }else{

        log_debug(ZONE,"SET %s",ns);

        /* compare the new data to the cache */
        if(cur->data != data)
        {
            /* the new data is different than the one in the cache, replace */
            xmlnode_free(cur->data);
            cur->data = data;
        }

        /* ask each xdb to update the user's data until one does */
        for(i = l; i != NULL; i = i->next)
            if((*(i->c))(1,cur,user,i->arg))
                break;
    }

    /* if we got here the data wasn't found, or the update was successful */
    return NULL;

}


/*
 *  js_xdb_register -- register a new user db with the server
 *  
 *  Creates a new xlist entry for this xdb module
 *
 *  parameters
 *  	c -- the module callback
 *		arg -- passed back to the callback
 */
void js_xdb_register(xcall c, void *arg)
{
    if(c == NULL) return;

    js_xdb(C_INIT, (udata)c, (char *)arg, NULL);
}


/*
 *  js_xdb_get -- retreive user data
 *  
 *  Searches each xdb module in turn for the
 *  requested user data.
 *
 *  parameters
 *  	udata -- the user for whose data we are searching
 *		ns -- a namespace indicating the type of data
 *            we're interested in. For eg. auth, roster, 
 *			  private storage etc.
 *
 *  returns
 *      the xmlnode requested, or NULL if not found 
 */
xmlnode js_xdb_get(udata user, char *ns)
{
    if(ns == NULL) return NULL;

    return js_xdb(C_GET, user, ns, NULL);
}

/*
 *  js_xdb_set -- store data for a user
 *  
 *  Stores an xml node and associates it with
 *  a paricular user. 
 *
 *  parameters
 *  	udata -- the user to whom the data belongs
 *		ns -- namespace for the data: auth, roster, private etc
 *		x -- the xmlnode to store
 *
 */
void js_xdb_set(udata user, char *ns, xmlnode x)
{
    if(ns == NULL) return;

    js_xdb(C_SET, user, ns, x);
}


